/* AP_AetherHarp-4_V60.
   Alex Pikkert October 2024.
   A music instrument with 8 notes in 6 octaves, played by moving your hands on a low (5 cm) or high (15 cm) level above 8 sensors.
   The tones are sent to a computer as digital codes, resulting in real music via the MIDI protocol, using 128 different instruments.
   The instrument is based on the Aetherharp design with a Teensy 3.5 computer board published by XenonJohn on the Instructables website:
   https://www.instructables.com/Aetherharp-Three-Level-24-Note-MIDI-Air-harp/
   My Aetherharp-4 specification can be found here: https://www.instructables.com/Aetherharp-4-MIDI-Air-Piano/
*/

//import libraries in the sketch

#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x27, 15, 2);
#include <SharpDistSensor.h>

#include <SimpleRotary.h>
// Rotary Pin A, Pin B, Button Pin
SimpleRotary rotary(1, 2, 3);

int Note[]        = {64, 66, 68, 69, 71, 73, 75, 76, 78, 81};
int AddNote       = 12;
bool NotePlayedLow[] = {false, false, false, false, false, false, false, false};
bool NotePlayedHigh[] = {false, false, false, false, false, false, false, false};
int SensorValue[] = {0, 0, 0, 0, 0, 0, 0, 0};
int SensorNumber = 8;
int BlueLed[] = {7, 9, 11, 24, 26, 28, 30, 32};
int RedLed[] = {6, 8, 10, 12, 25, 27, 29, 31};
const byte medianFilterWindowSize = 5;

// Define the array of SharpDistSensor objects
SharpDistSensor SensorArray[] = {
  SharpDistSensor(A14, medianFilterWindowSize),
  SharpDistSensor(A15, medianFilterWindowSize),
  SharpDistSensor(A16, medianFilterWindowSize),
  SharpDistSensor(A17, medianFilterWindowSize),
  SharpDistSensor(A18, medianFilterWindowSize),
  SharpDistSensor(A19, medianFilterWindowSize),
  SharpDistSensor(A20, medianFilterWindowSize),
  SharpDistSensor(A0, medianFilterWindowSize)
};

const int PitchBendPin = A6;
const byte upEncoderPin = 1;
const byte downEncoderPin = 2;
const byte InstrumentChangeButton = 3;
const byte SustainSwitch = 4;
const byte VibratoSwitch = 5;
const byte DuoModeSwitch = 21;
const byte OctaveDownButton = 22;
const byte OctaveUpButton = 23;
int Wait = 5;
int Octave = 2;
const char* OctString[5] = {"E2-1", "E3-2", "E4-3", "E5-4", "E6-5",};
bool SustainOn = false;
bool StaySelect = false;
bool DuoMode = false;
int h = 0; // rotary knob pushed/not pushed
int z = 0; // rotary knob longpush
int i = 0; // counter
int j = 0; // counter
int k = 0; // counter
int PitchBendSample;
int PitchValue;
int LowerTrigger;
int HigherTrigger;
int MidPoint;
int LedDelay = 100;

void setup() {
  //Enable serial communication for testing.
  // Serial.begin(115200);

  for (i = 0; i < SensorNumber; i++) {
    SensorArray[i].setModel(SharpDistSensor::GP2Y0A41SK0F_5V_DS);  // Set sensor model
  }

  //define the input pins with the built-in pullup resistors for the buttons & switches
  pinMode(1, INPUT_PULLUP);  // midi instrument selection clockwise
  pinMode(2, INPUT_PULLUP);  // midi instrument selection counterclockwise
  pinMode(3, INPUT_PULLUP);  // rotary encoder button for instrument reset
  pinMode(4, INPUT_PULLUP);  // midi select vibrato ON/OFF
  pinMode(5, INPUT_PULLUP);  // midi select sustain ON/OFF
  pinMode(21, INPUT_PULLUP); // midi select duomode ON/OFF
  pinMode(22, INPUT_PULLUP); // midi select octave DOWN
  pinMode(23, INPUT_PULLUP); // midi select octave UP
  pinMode(A6, INPUT);        // pitchbend input

  // define the output pins for 16 blue an red indication LEDs
  pinMode(6, OUTPUT);
  pinMode(7, OUTPUT);
  pinMode(8, OUTPUT);
  pinMode(9, OUTPUT);
  pinMode(10, OUTPUT);
  pinMode(11, OUTPUT);
  pinMode(12, OUTPUT);
  pinMode(24, OUTPUT);
  pinMode(25, OUTPUT);
  pinMode(26, OUTPUT);
  pinMode(27, OUTPUT);
  pinMode(28, OUTPUT);
  pinMode(29, OUTPUT);
  pinMode(30, OUTPUT);
  pinMode(31, OUTPUT);
  pinMode(32, OUTPUT);

  // Led test
  for (j = 0; j < SensorNumber; j++) {
    digitalWrite(BlueLed[j], HIGH);
    delay(LedDelay);
    digitalWrite(BlueLed[j], LOW);
  }
  delay(LedDelay);
  for (j = SensorNumber - 1; j >= 0; j--) {
    digitalWrite(RedLed[j], HIGH);
    delay(LedDelay);
    digitalWrite(RedLed[j], LOW);
  }
  delay(LedDelay);

  //setup for the start of the LCD display
  lcd.init();
  lcd.backlight();
  lcd.setCursor(0, 0); lcd.print(" *AETHERHARP-4* ");
  lcd.setCursor(0, 1); lcd.print("      *V60*     ");
  delay(3000);
  lcd.setCursor(0, 0); lcd.print("start MidiPlayer");
  lcd.setCursor(0, 1); lcd.print("                ");
  delay(3000);
  lcd.setCursor(0, 0); lcd.print("            E4-3");
  lcd.setCursor(0, 1); lcd.print("grand piano     ");

  //sets up an initial dead zone in the pitchbend LDRs either way of midpoint before a note is bent
  MidPoint = analogRead(PitchBendPin);
  HigherTrigger = MidPoint + 50;
  LowerTrigger = MidPoint - 50;

}

void loop() {
  // keep an eye on the switches and the pitchbend LDRs via subroutines
  CheckOctave();
  CheckVibrato();
  CheckDuoMode();
  CheckSustain();
  CheckPitchBend();
  CheckInstrumentChange();


  // ***********************DETECTING LOW HAND*******************************

  for (i = 0; i < SensorNumber; i++) {

    SensorValue[i] = SensorArray[i].getDist();
    delay(Wait);

    if (NotePlayedLow[i] == true && SustainOn == false && SensorValue[i]  < 8000)
    {
      digitalWrite(RedLed[i], LOW);
      usbMIDI.sendNoteOff(Note[i] - AddNote, 0, 1);
      if (DuoMode == true) {
        usbMIDI.sendNoteOff(Note[i + 2] - AddNote, 0, 1);
      }
      NotePlayedLow[i] = false;
    }

    if ( SensorValue[i] > 11000 && NotePlayedLow[i] == false)
    {
      digitalWrite(RedLed[i], HIGH);
      usbMIDI.sendNoteOn((Note[i] - AddNote), 255, 1);
      if (DuoMode == true) {
        usbMIDI.sendNoteOn(Note[i + 2] - AddNote, 255, 1);
        usbMIDI.sendNoteOff(Note[i + 2], 0, 1);
      }
      usbMIDI.sendNoteOff((Note[i] ), 255, 1);
      NotePlayedLow[i] = true;
    }
  }

  // ***********************DETECTING HIGH HAND*******************************

  for (i = 0; i < SensorNumber; i++) {

    SensorValue[i] = SensorArray[i].getDist();
    delay(Wait);

    if (NotePlayedHigh[i] == true && SustainOn == false && SensorValue[i]  < 1200)
    {
      digitalWrite(BlueLed[i], LOW);
      usbMIDI.sendNoteOff(Note[i], 0, 1);
      if (DuoMode == true) {
        usbMIDI.sendNoteOff(Note[i + 2], 0, 1);
      }
      NotePlayedHigh[i] = false;
    }

    if ( SensorValue[i] < 3000 && SensorValue[i] > 1200 && NotePlayedHigh[i] == false)
    {
      digitalWrite(BlueLed[i], HIGH);
      usbMIDI.sendNoteOn((Note[i] ), 255, 1);
      usbMIDI.sendNoteOff((Note[i] - AddNote), 0, 1);
      if (DuoMode == true) {
        usbMIDI.sendNoteOn(Note[i + 2], 255, 1);
        usbMIDI.sendNoteOff(Note[i + 2] - AddNote, 0, 1);
      }
      NotePlayedHigh[i] = true;
    }
  }
}

//***********************SUBROUTINES*************************************

//***********************************************************************
void CheckVibrato()
{
  //switch vibrato ON and show this on the display
  if (digitalRead(VibratoSwitch) == LOW ) {
    usbMIDI.sendControlChange(1, 127, 1);
    lcd.setCursor(0, 0);
    lcd.print("vib");
  }
  //switch vibrato OFF and show this on the display
  if (digitalRead(VibratoSwitch) == HIGH ) {
    usbMIDI.sendControlChange(1, 0, 1);
    lcd.setCursor(0, 0);
    lcd.print("   ");
  }
}

//***********************************************************************
void CheckDuoMode() {
  //switch duomode ON and show this on the display
  if (digitalRead(DuoModeSwitch) == LOW ) {
    DuoMode = true;
    lcd.setCursor(4, 0);
    lcd.print("duo");
  }
  //switch duomode OFF and show this on the display
  if (digitalRead(DuoModeSwitch) == HIGH ) {
    DuoMode = false;
    lcd.setCursor(4, 0);
    lcd.print("   ");
  }
}

//***********************************************************************
void CheckSustain()
{
  //switch sustain ON and show this on the display
  if (digitalRead(SustainSwitch) == LOW ) {
    usbMIDI.sendControlChange(64, 64, 1);
    SustainOn = true;
    lcd.setCursor(8, 0);
    lcd.print("sus");
  }
  //switch sustain OFF and show this on the display
  if (digitalRead(SustainSwitch) == HIGH ) {
    usbMIDI.sendControlChange(64, 0, 1);
    SustainOn = false;
    lcd.setCursor(8, 0);
    lcd.print("   ");
  }
}

//***********************************************************************
void CheckPitchBend()
{
  PitchBendSample = analogRead(PitchBendPin);  // 0 to 1023 in value
  PitchValue = 0;
  if (PitchBendSample < LowerTrigger) {
    PitchValue = (LowerTrigger - PitchBendSample) * (-50);
  }
  if (PitchBendSample > HigherTrigger) {
    PitchValue = (PitchBendSample - HigherTrigger) * 55;
  }
  if (PitchValue > 8192) {
    PitchValue = 8192;
  }
  if (PitchValue < -8192) {
    PitchValue = -8192;
  }
  usbMIDI.sendPitchBend(PitchValue, 1);
}

//***********************************************************************
void InstrumentReset() {

  lcd.setCursor(0, 1);
  lcd.print("grand piano     ");
  usbMIDI.sendProgramChange(0, 1);
  j = 1;
}

//***********************************************************************
void AllBlueLedsOn() {

  for (i = 0; i < SensorNumber; i++) {
    digitalWrite(BlueLed[i], HIGH);
  }
}

//***********************************************************************
void AllLedsOff() {

  for (i = 0; i < SensorNumber; i++) {
    digitalWrite(BlueLed[i], LOW);
    digitalWrite(RedLed[i], LOW);
  }
}

//***********************************************************************
void CheckOctave() {

  if (digitalRead(OctaveDownButton) == LOW) {
    delay(300);
    if (Octave > 0) {
      Octave = Octave - 1;
      for (k = 0; k < SensorNumber; k++) {
        usbMIDI.sendNoteOff(Note[k], 0, 1); Note[k] = Note[k] - AddNote;
      }
      lcd.setCursor(12, 0); 
      lcd.print(OctString[Octave]);
    }
  }

  if (digitalRead(OctaveUpButton) == LOW) {
    delay(300);
    if (Octave < 4) {
      Octave = Octave + 1;
      for (k = 0; k < SensorNumber; k++) {
        usbMIDI.sendNoteOff(Note[k], 0, 1); Note[k] = Note[k] + AddNote;
      }
      lcd.setCursor(12, 0); 
      lcd.print(OctString[Octave]);
    }
  }
}

//***********************************************************************
void CheckInstrumentChange() {

  // 0 = not pushed, 1 = pushed
  h = rotary.push();
  if ( h == 1 ) {
    AllLedsOff();
    StaySelect = true;
    delay(500);
  }
  while (StaySelect == true) {
    AllBlueLedsOn();
    lcd.setCursor(0, 0); lcd.print("*set instrument*");
    // 0 = not turning, 1 = CW, 2 = CCW
    h = rotary.rotate();

    if ( h == 1 ) {
      j = j + 1;
      delay(100);
      Select();
    }
    if ( h == 2 ) {
      j = j - 1;
      delay(100);
      Select();
    }
    // Check to see if button is pressed for 2 seconds.
    z = rotary.pushLong(2000);

    if ( z == 1 ) {
      InstrumentReset();
      StaySelect = false;
      AllLedsOff();
      lcd.setCursor(0, 0); lcd.print("                ");
      lcd.setCursor(12, 0); lcd.print(OctString[Octave]);
    }

    h = rotary.push();
    if ( h == 1 ) {
      StaySelect = false;
      AllLedsOff();
      lcd.setCursor(0, 0); lcd.print("                ");
      lcd.setCursor(12, 0); lcd.print(OctString[Octave]);
      delay(500);
    }
  }
}

//***********************************************************************
void Select()
{
  //this is the list of instruments, new instruments can be added
  if (j < 1  ) {
    j = 56;
  }
  if (j > 56 ) {
    j = 1;
  }

  switch (j) {

    case 1:
      lcd.setCursor(0, 1);
      lcd.print("grand piano     ");
      usbMIDI.sendProgramChange(0, 1);
      break;

    case 2:
      lcd.setCursor(0, 1);
      lcd.print("honky-tonk piano");
      usbMIDI.sendProgramChange(3, 1);
      break;

    case 3:
      lcd.setCursor(0, 1);
      lcd.print("harpsichord     ");
      usbMIDI.sendProgramChange(6, 1);
      break;

    case 4:
      lcd.setCursor(0, 1);
      lcd.print("glockenspiel    ");
      usbMIDI.sendProgramChange(9, 1);
      break;

    case 5:
      lcd.setCursor(0, 1);
      lcd.print("music box       ");
      usbMIDI.sendProgramChange(10, 1);
      break;

    case 6:
      lcd.setCursor(0, 1);
      lcd.print("vibraphone      ");
      usbMIDI.sendProgramChange(11, 1);
      break;

    case 7:
      lcd.setCursor(0, 1);
      lcd.print("xylophone       ");
      usbMIDI.sendProgramChange(13, 1);
      break;

    case 8:
      lcd.setCursor(0, 1);
      lcd.print("tubular bells   ");
      usbMIDI.sendProgramChange(14, 1);
      break;

    case 9:
      lcd.setCursor(0, 1);
      lcd.print("dulcimer       ");
      usbMIDI.sendProgramChange(15, 1);
      break;

    case 10:
      lcd.setCursor(0, 1);
      lcd.print("church organ    ");
      usbMIDI.sendProgramChange(19, 1);
      break;

    case 11:
      lcd.setCursor(0, 1);
      lcd.print("accordion       ");
      usbMIDI.sendProgramChange(21, 1);
      break;

    case 12:
      lcd.setCursor(0, 1);
      lcd.print("harmonica       ");
      usbMIDI.sendProgramChange(22, 1);
      break;

    case 13:
      lcd.setCursor(0, 1);
      lcd.print("tango accordion ");
      usbMIDI.sendProgramChange(23, 1);
      break;

    case 14:
      lcd.setCursor(0, 1);
      lcd.print("clean el. guitar");
      usbMIDI.sendProgramChange(27, 1);
      break;

    case 15:
      lcd.setCursor(0, 1);
      lcd.print("slap bass1      ");
      usbMIDI.sendProgramChange(36, 1);
      break;

    case 16:
      lcd.setCursor(0, 1);
      lcd.print("synth bass 2    ");
      usbMIDI.sendProgramChange(39, 1);
      break;

    case 17:
      lcd.setCursor(0, 1);
      lcd.print("violin          ");
      usbMIDI.sendProgramChange(40, 1);
      break;

    case 18:
      lcd.setCursor(0, 1);
      lcd.print("tremolo strings ");
      usbMIDI.sendProgramChange(44, 1);
      break;

    case 19:
      lcd.setCursor(0, 1);
      lcd.print("pizzicato string");
      usbMIDI.sendProgramChange(45, 1);
      break;

    case 20:
      lcd.setCursor(0, 1);
      lcd.print("orchestral harp ");
      usbMIDI.sendProgramChange(46, 1);
      break;

    case 21:
      lcd.setCursor(0, 1);
      lcd.print("synth strings 1 ");
      usbMIDI.sendProgramChange(50, 1);
      break;

    case 22:
      lcd.setCursor(0, 1);
      lcd.print("choir aahs      ");
      usbMIDI.sendProgramChange(52, 1);
      break;

    case 23:
      lcd.setCursor(0, 1);
      lcd.print("voice oohs      ");
      usbMIDI.sendProgramChange(53, 1);
      break;

    case 24:
      lcd.setCursor(0, 1);
      lcd.print("orchestra hit   ");
      usbMIDI.sendProgramChange(55, 1);
      break;

    case 25:
      lcd.setCursor(0, 1);
      lcd.print("trombone        ");
      usbMIDI.sendProgramChange(57, 1);
      break;

    case 26:
      lcd.setCursor(0, 1);
      lcd.print("french horn     ");
      usbMIDI.sendProgramChange(60, 1);
      break;
    case 27:
      lcd.setCursor(0, 1);
      lcd.print("alto sax        ");
      usbMIDI.sendProgramChange(65, 1);
      break;

    case 28:
      lcd.setCursor(0, 1);
      lcd.print("bariton sax     ");
      usbMIDI.sendProgramChange(67, 1);
      break;

    case 29:
      lcd.setCursor(0, 1);
      lcd.print("clarinet        ");
      usbMIDI.sendProgramChange(71, 1);
      break;

    case 30:
      lcd.setCursor(0, 1);
      lcd.print("flute           ");
      usbMIDI.sendProgramChange(73, 1);
      break;

    case 31:
      lcd.setCursor(0, 1);
      lcd.print("pan flute       ");
      usbMIDI.sendProgramChange(75, 1);
      break;

    case 32:
      lcd.setCursor(0, 1);
      lcd.print("blown bottle    ");
      usbMIDI.sendProgramChange(76, 1);
      break;

    case 33:
      lcd.setCursor(0, 1);
      lcd.print("sawtooth        ");
      usbMIDI.sendProgramChange(81, 1);
      break;

    case 34:
      lcd.setCursor(0, 1);
      lcd.print("fifths          ");
      usbMIDI.sendProgramChange(86, 1);
      break;

    case 35:
      lcd.setCursor(0, 1);
      lcd.print("bass + lead     ");
      usbMIDI.sendProgramChange(87, 1);
      break;

    case 36:
      lcd.setCursor(0, 1);
      lcd.print("new age         ");
      usbMIDI.sendProgramChange(88, 1);
      break;

    case 37:
      lcd.setCursor(0, 1);
      lcd.print("bowed           ");
      usbMIDI.sendProgramChange(92, 1);
      break;

    case 38:
      lcd.setCursor(0, 1);
      lcd.print("metallic        ");
      usbMIDI.sendProgramChange(93, 1);
      break;

    case 39:
      lcd.setCursor(0, 1);
      lcd.print("ice rain        ");
      usbMIDI.sendProgramChange(96, 1);
      break;

    case 40:
      lcd.setCursor(0, 1);
      lcd.print("soundtrack      ");
      usbMIDI.sendProgramChange(97, 1);
      break;

    case 41:
      lcd.setCursor(0, 1);
      lcd.print("crystal         ");
      usbMIDI.sendProgramChange(98, 1);
      break;

    case 42:
      lcd.setCursor(0, 1);
      lcd.print("echoes          ");
      usbMIDI.sendProgramChange(102, 1);
      break;

    case 43:
      lcd.setCursor(0, 1);
      lcd.print("sci-fi          ");
      usbMIDI.sendProgramChange(103, 1);
      break;

    case 44:
      lcd.setCursor(0, 1);
      lcd.print("sitar           ");
      usbMIDI.sendProgramChange(104, 1);
      break;

    case 45:
      lcd.setCursor(0, 1);
      lcd.print("banjo           ");
      usbMIDI.sendProgramChange(105, 1);
      break;

    case 46:
      lcd.setCursor(0, 1);
      lcd.print("shamisen        ");
      usbMIDI.sendProgramChange(106, 1);
      break;

    case 47:
      lcd.setCursor(0, 1);
      lcd.print("koto            ");
      usbMIDI.sendProgramChange(107, 1);
      break;

    case 48:
      lcd.setCursor(0, 1);
      lcd.print("kalimba         ");
      usbMIDI.sendProgramChange(108, 1);
      break;

    case 49:
      lcd.setCursor(0, 1);
      lcd.print("shanai          ");
      usbMIDI.sendProgramChange(111, 1);
      break;

    case 50:
      lcd.setCursor(0, 1);
      lcd.print("tinkle bell     ");
      usbMIDI.sendProgramChange(112, 1);
      break;

    case 51:
      lcd.setCursor(0, 1);
      lcd.print("agogo           ");
      usbMIDI.sendProgramChange(113, 1);
      break;

    case 52:
      lcd.setCursor(0, 1);
      lcd.print("steel drums     ");
      usbMIDI.sendProgramChange(114, 1);
      break;

    case 53:
      lcd.setCursor(0, 1);
      lcd.print("taiko drum      ");
      usbMIDI.sendProgramChange(116, 1);
      break;

    case 54:
      lcd.setCursor(0, 1);
      lcd.print("guitarfret noise");
      usbMIDI.sendProgramChange(120, 1);
      break;

    case 55:
      lcd.setCursor(0, 1);
      lcd.print("seashore        ");
      usbMIDI.sendProgramChange(122, 1);
      break;

    case 56:
      lcd.setCursor(0, 1);
      lcd.print("bird tweet      ");
      usbMIDI.sendProgramChange(123, 1);
      break;

    default:
      break;
  }
}

//****************************END OF SKETCH******************************
